// Author: A.Voss@FH-Aachen.de
//
// mutex usage
// tags: thread, mutex

#include <iostream>
#include <thread>
#include <condition_variable>
#include <chrono>
#include <string>
#include <vector>
#include <algorithm>
using namespace std;

#include "../../include/tools.h"

// thread-function, version 1, unsafe
void Wrk1(long long inc, long long n, long long * sum) 
{
    tools_log(); cout << "thread started, start loop version 1" << endl;

    for (long long i=0; i<n; ++i) 
    {
        (*sum) += inc;
    }  

    tools_log(); cout << "loop and thread done" << endl;     
}

// thread-function, version 2, safe
void Wrk2(long long inc, long long n, long long * sum, mutex * mutex_sum) 
{
    tools_log(); cout << "thread started, start loop version 2" << endl;

    for (long long i=0; i<n; ++i) 
    {
        mutex_sum->lock();
        (*sum) += inc;      // note: in general, this solution is unsafe -> why?
        // v.push_back(n)   // hint: what may happen here
        mutex_sum->unlock();
    }  

    tools_log(); cout << "loop and thread done" << endl;     
}

int main()
{
    const long long n = 10000000;
    long long sum;

    tools_log(); cout << "start threads t1 and t2 and sum" << endl;

    sum = 0;
    thread t1(Wrk1,+1,n,&sum); tools_registerthreadname(t1); 
    thread t2(Wrk1,-1,n,&sum); tools_registerthreadname(t2); 
    t1.join();
    t2.join();
    tools_log(); cout << "threads joined, sum=" << sum << endl;

    tools_log(); cout << "start threads t3 and t4 and sum" << endl;

    mutex mutex_sum;
    
    sum = 0;
    thread t3(Wrk2,+1,n,&sum,&mutex_sum); tools_registerthreadname(t3); 
    thread t4(Wrk2,-1,n,&sum,&mutex_sum); tools_registerthreadname(t4); 
    t3.join();
    t4.join();
    tools_log(); cout << "threads joined, sum=" << sum << endl;
    
    return EXIT_SUCCESS;
}
